use std::alloc::{alloc, dealloc, Layout};
use std::marker::PhantomData;
use std::ptr::{copy_nonoverlapping, NonNull};
use std::slice::{from_raw_parts, from_raw_parts_mut};

/*
 * Define supporting types, functions, and macros for Hercules RT functions. For
 * a more in-depth discussion of the design of these utilities, see hercules_cg/
 * src/rt.rs (the RT backend).
 */

pub unsafe fn __cpu_alloc(size: usize) -> *mut u8 {
    alloc(Layout::from_size_align(size, 16).unwrap())
}

pub unsafe fn __cpu_dealloc(ptr: *mut u8, size: usize) {
    dealloc(ptr, Layout::from_size_align(size, 16).unwrap())
}

pub unsafe fn __copy_cpu_to_cpu(dst: *mut u8, src: *mut u8, size: usize) {
    copy_nonoverlapping(src, dst, size);
}

#[cfg(feature = "cuda")]
extern "C" {
    pub fn __cuda_alloc(size: usize) -> *mut u8;
    pub fn __cuda_dealloc(ptr: *mut u8, size: usize);
    pub fn __copy_cpu_to_cuda(dst: *mut u8, src: *mut u8, size: usize);
    pub fn __copy_cuda_to_cpu(dst: *mut u8, src: *mut u8, size: usize);
    pub fn __copy_cuda_to_cuda(dst: *mut u8, src: *mut u8, size: usize);
}

#[derive(Clone, Debug)]
pub struct HerculesCPURef<'a> {
    ptr: NonNull<u8>,
    size: usize,
    _phantom: PhantomData<&'a u8>,
}

#[derive(Debug)]
pub struct HerculesCPURefMut<'a> {
    ptr: NonNull<u8>,
    size: usize,
    _phantom: PhantomData<&'a u8>,
}

#[cfg(feature = "cuda")]
#[derive(Clone, Debug)]
pub struct HerculesCUDARef<'a> {
    ptr: NonNull<u8>,
    size: usize,
    _phantom: PhantomData<&'a u8>,
}

#[cfg(feature = "cuda")]
#[derive(Debug)]
pub struct HerculesCUDARefMut<'a> {
    ptr: NonNull<u8>,
    size: usize,
    _phantom: PhantomData<&'a u8>,
}

#[cfg(feature = "cuda")]
#[derive(Debug)]
pub struct CUDABox {
    ptr: NonNull<u8>,
    size: usize,
}

impl<'a> HerculesCPURef<'a> {
    pub fn from_slice<T>(slice: &'a [T]) -> Self {
        let ptr = unsafe { NonNull::new_unchecked(slice.as_ptr() as *mut u8) };
        let size = slice.len() * size_of::<T>();
        Self {
            ptr,
            size,
            _phantom: PhantomData,
        }
    }

    pub fn as_slice<T>(self) -> &'a [T] {
        let ptr = self.ptr.as_ptr() as *const T;
        assert_eq!(self.size % size_of::<T>(), 0);
        assert!(ptr.is_aligned());
        unsafe { from_raw_parts(ptr, self.size / size_of::<T>()) }
    }

    pub unsafe fn __ptr(&self) -> *mut u8 {
        self.ptr.as_ptr() as *mut u8
    }

    pub unsafe fn __size(&self) -> usize {
        self.size
    }

    pub unsafe fn __from_parts(ptr: *mut u8, size: usize) -> Self {
        Self {
            ptr: NonNull::new(ptr).unwrap(),
            size,
            _phantom: PhantomData,
        }
    }
}

impl<'a> HerculesCPURefMut<'a> {
    pub fn from_slice<T>(slice: &'a mut [T]) -> Self {
        let ptr = unsafe { NonNull::new_unchecked(slice.as_ptr() as *mut u8) };
        let size = slice.len() * size_of::<T>();
        Self {
            ptr,
            size,
            _phantom: PhantomData,
        }
    }

    pub fn as_slice<T>(self) -> &'a mut [T] {
        let ptr = self.ptr.as_ptr() as *mut T;
        assert_eq!(self.size % size_of::<T>(), 0);
        assert!(ptr.is_aligned());
        unsafe { from_raw_parts_mut(ptr, self.size / size_of::<T>()) }
    }

    pub fn as_ref(self) -> HerculesCPURef<'a> {
        HerculesCPURef {
            ptr: self.ptr,
            size: self.size,
            _phantom: PhantomData,
        }
    }

    pub unsafe fn __ptr(&self) -> *mut u8 {
        self.ptr.as_ptr()
    }

    pub unsafe fn __size(&self) -> usize {
        self.size
    }

    pub unsafe fn __from_parts(ptr: *mut u8, size: usize) -> Self {
        Self {
            ptr: NonNull::new(ptr).unwrap(),
            size,
            _phantom: PhantomData,
        }
    }
}

#[cfg(feature = "cuda")]
impl<'a> HerculesCUDARef<'a> {
    pub unsafe fn __ptr(&self) -> *mut u8 {
        self.ptr.as_ptr()
    }

    pub unsafe fn __size(&self) -> usize {
        self.size
    }

    pub unsafe fn __from_parts(ptr: *mut u8, size: usize) -> Self {
        Self {
            ptr: NonNull::new(ptr).unwrap(),
            size,
            _phantom: PhantomData,
        }
    }
}

#[cfg(feature = "cuda")]
impl<'a> HerculesCUDARefMut<'a> {
    pub fn as_ref(self) -> HerculesCUDARef<'a> {
        HerculesCUDARef {
            ptr: self.ptr,
            size: self.size,
            _phantom: PhantomData,
        }
    }

    pub unsafe fn __ptr(&self) -> *mut u8 {
        self.ptr.as_ptr()
    }

    pub unsafe fn __size(&self) -> usize {
        self.size
    }

    pub unsafe fn __from_parts(ptr: *mut u8, size: usize) -> Self {
        Self {
            ptr: NonNull::new(ptr).unwrap(),
            size,
            _phantom: PhantomData,
        }
    }
}

#[cfg(feature = "cuda")]
impl CUDABox {
    pub fn from_cpu_ref(cpu_ref: HerculesCPURef) -> Self {
        unsafe {
            let size = cpu_ref.size;
            let ptr = NonNull::new(__cuda_alloc(size)).unwrap();
            __copy_cpu_to_cuda(ptr.as_ptr(), cpu_ref.ptr.as_ptr(), size);
            Self { ptr, size }
        }
    }

    pub fn from_cuda_ref(cuda_ref: HerculesCUDARef) -> Self {
        unsafe {
            let size = cuda_ref.size;
            let ptr = NonNull::new(__cuda_alloc(size)).unwrap();
            __copy_cuda_to_cuda(ptr.as_ptr(), cuda_ref.ptr.as_ptr(), size);
            Self { ptr, size }
        }
    }

    pub fn get_ref<'a>(&'a self) -> HerculesCUDARef<'a> {
        HerculesCUDARef {
            ptr: self.ptr,
            size: self.size,
            _phantom: PhantomData,
        }
    }

    pub fn get_ref_mut<'a>(&'a mut self) -> HerculesCUDARefMut<'a> {
        HerculesCUDARefMut {
            ptr: self.ptr,
            size: self.size,
            _phantom: PhantomData,
        }
    }
}

#[cfg(feature = "cuda")]
impl Clone for CUDABox {
    fn clone(&self) -> Self {
        Self::from_cuda_ref(self.get_ref())
    }
}

#[cfg(feature = "cuda")]
impl Drop for CUDABox {
    fn drop(&mut self) {
        unsafe {
            __cuda_dealloc(self.ptr.as_ptr(), self.size);
        }
    }
}

#[macro_export]
macro_rules! runner {
    ($x: ident) => {
        <concat_idents!(HerculesRunner_, $x)>::new()
    };
}
